''************************************
''*         Pulse Generator V2.0     *
''************************************

''* 
''*  ---------------------------------------------------------  *
''*  Version 2.0
''*
''*  Corrected Error. In free-running mode, if an odd # of
''*  pulses was specified, an unending pulse stream was generated.
''*  Alterate output pin code following the initial pulse
''*  count loop would only set a_mode = 0 (preventing the
''*  pulse pattern code from being called repeatedly) for only
''*  one of the two alternate pin possibilities.

''* 
''*  ---------------------------------------------------------  *
''*  Version 1.0
''*
''*  Two-pin time/pin-triggerable pulse generator
''* 
''*  Methods:
''*     .Start(OutputPin, TriggerPin)
''*     .Stop
''*     .Mode() : 0 thru 4
''*     .Set

CON

'Number of hub variables of each type
  CogCnt = 2 
  ParaCnt = 8
  PinCnt = 2
  
  TestPin = 7   'useful during development if a scope is handy


OBJ

  'debugging only 
  'serial       : "SerialMirror2"             '1 Cog 
  
VAR

'Arbitrary convention: leading underscore for all hub variables

  long  _cog        '   Cog flag/id

 '----- Cog control: CogCnt Longs -------
   
  long  _command    '   If == 0, No command to processing (default).
                    '            The cog resets _command to zero at completion of command processing.
                    '   If == 1, Cog will copy the object's parameter list from hub-to-cog.
                    '            This will always force PulseSetup == 0, reinitializing the pulser.
                    '   If == 2, Cog will return the value of parameter defined by _index
                    '            In addition to parameters, a few read-only variables following the
                    '            parameter list may also be accessed.  

  'Add additional control variables here.
  'Assembly code assumes '_index' always immediately precedes the 1st parameter in parameter list.

  long  _index      '   The one-based index (1..ParaCnt) of parameter to be read back.
                    '   Upon completion of _command == 2, _index holds value of indexed parameter
                    
 '----- Parameters: ParaCnt Longs -------
 
  long  _mode       'p1      Trigger mode: 0 = pulse generator disabled (default) 
                    '                      1 = free-running
                    '                      2 = trigger, low-to-high
                    '                      3 = trigger, high-to-low
                    '                      4 = trigger, at '_sync' time
  long  _delay      'p2      Delay from trigger leading edge to 1st pulse leading edge
  long  _width      'p3      Width of each pulse
  long  _period     'p4      Period between pulses
  long  _number     'p5      Number of pulses per trigger event
  long  _sync       'p6      Cnt == _sync triggers pulse pattern output      
  long  _AltPin     'p7      Default: -1, else enables & specifies pin# to become an
                    '        an output pin which toggles at each pulse leading-edge.
  long  _GatePin    'p8      Default: -1, else enables & specifies pin# which must
                    '        be HIGH before trigger pulse will be sensed ...
                                        
 '------ Pins: PinCnt Longs -----    

  long  _OutPin     'pin#                
  long  _TrigPin    'pin# 
  
 '------ Hub Only Variables (not copied to cog) -------

  long  _units      
  long  _clocks_per_uS
   
PUB start( OutputPin, TriggerPin) : okay
                             
 'method call - stop already running cog ..
  stop

 'initialize local hub variables to zero 
  longfill(@_command, 0, CogCnt + ParaCnt)

  _units := 0  'default to units of microseconds

  _clocks_per_uS := clkfreq/1_000_000  'clockticks per microsecond

  _AltPin := -1  'by default disabled
  _GatePin := -1 'by default disabled
            
 'copy .start method's two parameters to local hub variables
  longmove(@_OutPin, @OutputPin, 2)
   
 'Note: par = hub_var0 within assembly routine
 '      cog = ID# of cog started + 1
 '      okay = return value for start() routine,
 '      cognew return -1 if no cog available,
 '      so after +1, cog =0 (FALSE), okay = FALSE
  okay := _cog := cognew(@entry, @_command) + 1

  
PUB stop

'' Stop serial driver - frees a cog

  if _cog                   'TRUE = non-zero
    cogstop(_cog~ - 1)      'stop cog with ID# 'cog-1', then post-clear cog to zero (FALSE)
    
  longfill(@_command, 0, CogCnt + ParaCnt)
      'write longs with zero value, starting at hub location '_command'.


PUB Read_Parameter(Pindex): Pvalue

    _index := Pindex           '_index holds 1-based index to parameter list prior to command.

    _command := 2              'command cog to return the value of the indexed parameter
    Repeat until _command == 0
    
    Pvalue := _index           '_index holds value of index parameter at command completion
                                
PUB Write_Parameters

    _command := 1               'command cog to copy its parameter list from the hub
    Repeat until _command == 0  
        
PUB Mode( value)

    'Trigger mode:
    '-------------
    '   0 = disabled (default) 
    '   1 = free-running mode
    '   2 = trigger on low-to-high
    
    _mode := value

PUB ToggleOutputEnable( TogglePin )

    _AltPin := TogglePin

PUB GateEnable (GatePin)

    _GatePin := GatePin 
     
PUB Period( value)

    _period := ConvertToClockTicks(value)

PUB Delay (value)

    _delay := ConvertToClockTicks(value) - 12  'correction for finite delay from pin change detection
                                               'to sampling of system count
     
PUB Width (value)

    _width := ConvertToClockTicks(value)

PUB Number (value)

    _number := value
        
PUB Sync (value)

    _sync := value

PUB Units (value)

    _units := value

PUB RemainingPulses : pulses_remaining

    if (_mode == 1) 
      pulses_remaining :=  Read_Parameter(ParaCnt + 1) 
    else
      pulses_remaining := 0

PRI ConvertToClockTicks(unit_val) : clockticks

   Case _units
      0:   'from microseconds to clock ticks 
         clockticks := unit_val * _clocks_per_uS
      1:   'from tenths of microseconds to clock ticks
         clockticks := unit_val * _clocks_per_uS / 10
      other:   'system clock ticks (raw, with no offset corrections)
         clockticks := unit_val

PUB ShowVariables

{
   serial.str(string(13,"Pulse Generator Variables: ",13))
   
   serial.str(string("pg_Mode = "))
   serial.dec(_mode)
   serial.putc(13)

   serial.str(string("pg_Delay = "))
   serial.dec(_delay)
   serial.putc(13)

   serial.str(string("pg_Width = "))
   serial.dec(_width)
   serial.putc(13)

   serial.str(string("pg_Period = "))
   serial.dec(_period)
   serial.putc(13)
     
   serial.str(string("pg_Number = "))
   serial.dec(_number)
   serial.putc(13)

   serial.str(string("pg_Sync = "))
   serial.dec(_sync)
   serial.putc(13)

   serial.str(string("pg_AltPin = "))
   serial.dec(_AltPin)
   serial.putc(13)

   serial.str(string("pg_GatePin = "))
   serial.dec(_GatePin)
   serial.putc(13)
   
}

DAT

'***********************************
'* Assembly language pulse driver  *
'***********************************
'
' Note: PAR holds the address of _command, a hub variable.
'       It was assigned by the COGNEW spin instruction within the .Start method.

'----- COG CONTROL -----

'par + 0 ==  _command  '   If == 0, No command to process (default).
'                                   The cog resets _command to zero at completion of cmd processing.
'                      '   If == 1, cog will copy the object's parameter list from hub-to-cog.
'                      '   If == 2, cog will return the value of a parameter defined by _index  

'par + 1 ==  _index    '   The one-based index (1..ParaCnt) of parameter to be read back.
                       '   Upon completion of _command == 2, _index holds value of indexed parameter.
                    
'----- PARAMETER LIST ----    ( 1st parameter's '_command' offset: Par + CogCnt )
 
'par + 2 ==  _mode     'p1      Trigger mode: 0 = pulse generator disabled (default) 
'                                 1 = free-running
'                                 2 = trigger, low-to-high
'                                 3 = trigger, high-to-low
'                                 4 = trigger, at '_sync' time
'par + 3 ==  _delay    'p2      Delay from trigger leading edge to 1st pulse leading edge
'par + 4 ==  _width    'p3      Width of each pulse
'par + 5 ==  _period   'p4      Period between pulses
'par + 6 ==  _number   'p5      Number of pulses per trigger event
'par + 7 ==  _sync     'p6      Cnt == _sync triggers pulse pattern output      
                      
 '----- PIN ASSIGNMENTS ----  ( 1st pin's '_command' offset: Par + CogCnt + ParaCnt )   

'par + 8 ==  _OutPin               
'par + 9 ==  _TrigPin     


'#########################################################################################

'  The ENTRY portion of this assembly language program accesses
'  a string of consecutive long variables, beginning at the hub
'  location, _COMMAND, using the special register, PAR.
'  ENTRY code creates the pin masks, and outputs default logic levels.


'NOTE:  "<<2" (below) multiplies "#N" by 4 (the number of bytes per long-type variable).
'       Hub memory is index by byte, while cog memory is always indexed by longs !!

                        org

entry                   mov     Initialized, #0 'FALSE
                        mov     PulseSetup,   #0
                                                
                        mov     a1,par         'get address of 1st long in hub parameter list
                        add     a1,#(CogCnt+ParaCnt) << 2  'skip to pin# assignments ...
                                               '(ADVANCE 'CogCnt + ParaCnt' LONGS)
                                                      
                        rdlong  d1,a1          'get _OutPin data value from hub memory
                        mov     OutMask,#1
                        shl     OutMask,d1     'OutMask:
                                               ' only set bit corresponding to _OutPin# position 
                        
                        
                        add     a1,#4          'get _TrigPin address  (ADVANCE 1 LONG)
                        rdlong  d1,a1          'get _TrigPin data value from hub memory
                        mov     TrigMask,#1      
                        shl     TrigMask,d1    'TrigMask:
                                               ' only set bit corresponding to _TrigPin# position 

                        mov     outa, #0       'By default, all cog pins used as outputs: LOW
                                               'This will override another cog's outputs
                                               'only if identical pins are assigned (by mistake).

                        mov     t1, OutMask
                        mov     dira, t1       'Make _OutPin an output, _TrigPin an input
                                               'TestPin an output
                                               
'---------------------------------------------------------------------------
'----------------------------   MAIN CODE LOOP  ----------------------------
'---------------------------------------------------------------------------

MainLoop               call   #CommandTest  'has cog received a command ?
                              
                      '.... don't generate pulses until after first write_parameters() call !
                       tjz  Initialized, #MainLoop

                      '.... don't generate pulses if _mode == 0 (pulse generator disabled)  
                       tjz  a_mode, #MainLoop

                       cmp  a_mode, #1    wz
               if_Z    call   #PulsePattern1   'Free running mode

                       cmp  a_mode, #2    wz
               if_Z    call   #PulsePattern2   'Trigger on L-to-H mode     
                
                       jmp  #MainLoop
                       
                       
'---------------------------------------------------------------------------
'-------------------------  Test for command assertion ---------------------
'---------------------------------------------------------------------------

CommandTest             mov     a1,par
                        rdlong  d1,a1     wz    'read _command (in hub memory)                        
                                               
         if_Z           jmp  CommandTest_RET    'if _command == 0 then exit (no command asserted)
                                                'NOTE: without the "#" literal, this jmp instruction
                                                'is actually reading the low 9-bits of the RET
                                                'instruction at location CommandTest_RET. Faster !

                        djnz    d1, #READ_PARAM '_command == 1 (UPDATE PARAMETERS) ??

                        
'---------------------------------------------------------------------------
'------------     Update cog with hub parameters values  -------------------
'---------------------------------------------------------------------------

UPDATE_PARAMETERS       mov     c1,#ParaCnt+1  'ParaCnt+1 hub variables to be copied to cog
                                               '_index plus each hub parameter will be copied

                        mov     a2, #a_index   'Preload a2 with 1st cog address in sequential list
                                               'of cog 'longs' to be filled with hub memory values.

                        mov     a1, par            'Get '_command' hub address
                        add     a1, #(CogCnt-2)<<2 'advance hub address pointer to _index-1
                                                   '_index is assumed to always be located just
                                                   'prior to first parameter in hub parameter list.      
                                                   'This ADD instruction makes code CogCnt-independent.
'----                                               
LoopX                   add     a1, #4         'advance hub address pointer by 1 Long      
                        rdlong  d1,a1          'fetch content of hub var at addr a1, write to cog loc d1

                        
              '--- modify next mov op, changing its destination addr to the value held by cog var, a2 

                        movd    IndirectWrite, a2
                        nop
IndirectWrite           mov     xx,d1          'Write hub data to cog address given by value of a2
                                               'NOTE: Dest. reg xx was changed by the earlier
                                               '      MOVD instruction to the value of a2( a cog address)
                                               
                        add     a2, #1         'Advance to next reserved long in cog memory
                                               'CRITICAL: cog is long-based !!      
                        djnz    c1, #LoopX

              '--- signal completion of first parameter update (Initialized ==0 following .start)
                             
                        mov     Initialized, cog_true   'Set Initialized := TRUE
                        mov     PulseSetup, #1          'Force PulsePattern Setup

                        jmp     #COMMAND_DONE

'-------------------------------------------------------------------------------                                                 
'---- Return value of cog parameter whose list index = a_index (one-based)  ----
'-------------------------------------------------------------------------------
              
READ_PARAM              djnz    d1, #COMMAND_DONE ' _command == 2 (READ_PARAMETER) ??


                      'Get value of hub variable _index (index to parameter list) 

                        mov     a1, par            'get address of _command
                        add     a1, #(CogCnt-1)<<2 'a1 == address of _index (just before para list)        
                                                   'just prior to hub parameter list
                        rdlong  d1, a1             'read current value of _index from hub memory

                      'Set cog address (of wrlong instruction) to the indexed parameter's location within cog
                                                   
                        mov     a2, #a_index     'Cog address of 1st parameter in list MINUS 1
                        add     a2, d1           'd1 == one-based index into parameter list
                        movd    ParaWrite, a2    'ParaWrite's instruction (wrlong) dest field
                                                 'is set to the value of a2 (the cog address from 
                                                 'which data is to be retrieved for sending to hub)
                                                 
                        nop
ParaWrite               wrlong  xx, a1          'write to _index (hub var) the value of indexed parameter  
                       
                        jmp     #COMMAND_DONE

'----------------------------------------------------------------------------------                        
'---- Signal completion of command operation by resetting _command := FALSE(0) ----
'----------------------------------------------------------------------------------
 
COMMAND_DONE            mov     a1,par           'address of '_command' hub variable
                        wrlong  cog_false, a1    'set _command :== FALSE (0)
                                                                                    
CommandTest_RET         ret


'-----------------------------------------------------------------------------
'-------------------- Pulse pattern 1 : Free-running mode --------------------
'-----------------------------------------------------------------------------
 
PulsePattern1           tjz      PulseSetup, #TrigPulse1 

                        mov      PulseSetup, #0    'Clear PulseSetup flag 

                        mov      PulseCounter, a_number

'------ Test a_altpin's value to see if it is >= 0. If so, modify Out_LeadEdge1
'       to generate an alternating logic level at the specified alternate pin output.
                  
                        abs      AltFlag, a_altpin  WC, NR 'if a_altpin == negative, C==1

                        mov      AltFlag, #0            'reset to zero (alt pin == LOW)
                        mov      Out_LeadEdge0, OutMask 'Only Out_pin HIGH
                        mov      Out_LeadEdge1, OutMask 'Modified if AltPin Enabled
                                                        'so that Alt Pin is also HIGH
                                                        
               IF_C     jmp      #PeriodTest    'skip as alternate output pin is disabled
                  
                        mov      AltMask,#1
                        shl      AltMask, a_altpin       'Create AltMask
                        or       Out_LeadEdge1, AltMask  'Out & Alt pins both HIGH  
                        or       dira, AltMask           'Define Alt pin as an output
                                             
PeriodTest              mov      To_end_of_period, a_period
                        sub      To_end_of_period, a_width   WC 'test for underflow

               IF_C     mov      a_mode, #0          'Abort pulser action if a_width > a_period !!
               IF_C     jmp      PulsePattern1_RET   '#' is not required. Indirect jump is faster.

'----------- Pulse Pattern Trigger ---------

TrigPulse1              mov      SyncPoint, cnt      'Capture SyncPoint
                        add      SyncPoint, #100     'Offset 100 clock ticks into future ...
    
'------------ Pulse Repeat Loop Entry ---------

NextPulse1              tjz      AltFlag, #AltPulse0 
 
'------------ Alternate Output Pin HIGH -----------------------------

AltPulse1               waitcnt  SyncPoint, a_width   'Wait for cnt == SyncPoint
                                                      'New SyncPoint == SyncPoint + a_width
                        mov      outa, Out_LeadEdge1  'Set Output Pin HIGH                      
                                                      'Set Alternate Pin HIGH
                                                      
                        waitcnt  SyncPoint, To_end_of_period  'Wait for cnt == SyncPoint
                                                      'New SyncPoint == SyncPoint + (a_period-a_width) 
                        andn     outa, OutMask        'Set Output Pin LOW
                                                              
                        jmp      #AfterPulse

'------------ Alternate Output Pin LOW -----------------------------

AltPulse0               waitcnt  SyncPoint, a_width   'Wait for cnt == SyncPoint
                                                      'New SyncPoint == SyncPoint + a_width
                        mov      outa, Out_LeadEdge0  'Set Output Pin HIGH                      
                                                      'Set Alternate Pin LOW
                      
                        waitcnt  SyncPoint, To_end_of_period  'Wait for cnt == SyncPoint
                                                      'New SyncPoint == SyncPoint + (a_period-a_width) 
                        andn     outa, OutMask        'Set Output Pin LOW
      
'----------  Immediately following each pulse's trailing-edge test for a new command.

AfterPulse              xor    AltFlag, cog_true      'toggle all 32-bits of flag after each pulse
                                                  
                        call   #CommandTest 'new command ?
          
                       'Note: PulseSetup == 0, unless new parameters have just been loaded,
                        tjnz PulseSetup, PulsePattern1_RET  
                       'EXIT if new parameters have been loaded !
                       'Note: "#" not used (for speed sake)
                         
                IF_C    mov      PulseCounter, #0    'SPECIAL CASE : by re-zeroing PulseCounter the pulser
                                                     'will run forever (provided a_number == 0 ). Carry is
                                                     'set by the djnz (wc) instruction at loop's end.
               
                        djnz PulseCounter, #NextPulse1  WC    'loop
                    
    'NOTE: if a_number == 0, PulseCounter is initially set to 0 as well.
    'The DJNZ result at 1st decrement will be < zero, setting the carry flag (WC).
    'Note that the DJNZ zero test occurs AFTER the decrement operation, so the loop is not exited.
    'The "IF_C mov PulseCounter, #0" statement when encountered next time through
    'the loop will reset PulseCounter back to zero, causing the same DJNZ result
    'so that looping (and output pulses) continues indefinitely.

                       
'---------- Following last pulse wait until end of period before
'           performing a final toggle of the alternate pin 
                        
                        tjz      AltFlag, #AltEnd0
                   
                        waitcnt  SyncPoint, #0       'Wait for end of period ...
                        or       outa, AltMask       'Set Alt Pin == HIGH

                        mov a_mode, #0   'disable pulser following last pulse in sequence
                    
                        jmp      PulsePattern1_RET
                        
AltEnd0                 waitcnt  SyncPoint, #0       'Wait for end of period ...
                        andn     outa, AltMask       'Set Alt Pin == LOW
                                                 
                        mov a_mode, #0   'disable pulser following last pulse in sequence
                    
                                   
PulsePattern1_RET       ret

'---------------------------------------------------------------------------------
'                     USEFUL TOOL FOR DEBUGGING PURPOSES
'---------------------------------------------------------------------------------
TestPulse               or      dira, TestMask   'At first call make '' pin#                                             
                                                 'an output
                                                 
                        or      outa, TestMask   'Set TestPin HIGH  -debug only
                        nop
                        nop
                        nop
                        nop
                        nop
                        nop
                        andn    outa, TestMask   'Set TestPin LOW  - debug only
TestPulse_RET           ret


'-----------------------------------------------------------------------------
'------------------ Pulse pattern 2 : Triggered Low-To_High ------------------
'-----------------------------------------------------------------------------
  
PulsePattern2           tjz      PulseSetup, #TrigPulse2 

                        mov      PulseSetup, #0    'Lower flag following sync operation

                       'Test for a_gatePin begin non-negative
                       'if so, we must modify TrigMask so that it includes the
                       'requirement that the Gate pin is HIGH in addition to a
                       'low to high Trig pin transition. 

                        mov      TrigGateMask, TrigMask  'preload with default non-gate value                 
                        abs      t1, a_gatePin  WC, NR 'if a_gatePin == negative, C==1
               IF_C     jmp      #PeriodTest2          'skip if gate disabled

                        mov      GateMask,#1
                        shl      GateMask, a_GatePin      'Create GateMask
                        or       TrigGateMask, GateMask
                        
PeriodTest2             mov      To_end_of_period, a_period
                        sub      To_end_of_period, a_width    WC  'test for underflow
                        
               IF_C     mov      a_mode, #0          'Abort pulser action if a_width > a_period !!
               IF_C     jmp      PulsePattern2_RET   '#' is not required. Indirect jump is faster.
                        
'---------
TrigPulse2              mov      PulseCounter, a_number
                        mov      SyncPoint, a_delay  'Offset in clock ticks to leading-edge of 1st pulse  

                        waitpne  TrigMask, TrigMask          'Wait for TrigPin  = LOW 
                        waitpeq  TrigGateMask, TrigGateMask  'Wait for TrigPin = HIGH
                                                         '(and if gated, GatePin = HIGH)
                                                         
                        add      SyncPoint, cnt      'Update SyncPoint IMMEDIATELY following trigger  !!!
                     
'---------                    
NextPulse2              waitcnt  SyncPoint, a_width 'Wait for cnt == SyncPoint
                                                    'New SyncPoint == SyncPoint + a_width
                        or       outa, OutMask      'Set Output Pin HIGH                      
                      
                      
                        waitcnt  SyncPoint, To_end_of_period  'Wait for cnt == SyncPoint
                                                    'New SyncPoint == SyncPoint + (a_period - a_width) 
                        andn     outa, OutMask      'Set Output Pin LOW
                     

                       'Immediately following each pulse's trailing-edge test for a new command.
                        call   #CommandTest  'new command issued ?

                       'PulseSetup == 0, unless new parameters have just been loaded,
                        tjnz PulseSetup, PulsePattern2_RET  'EXIT if new parameters have been loaded. 
                                                           'note: "#" not used (for speed sake)

                        djnz PulseCounter, #NextPulse2  WC    'loop 
                                                        'NOTE: if a_number == 0 the decrement result will
                                                        'be non-zero and pulser action will continue.

                        jmp #TrigPulse2  'wait for next trigger event  
            
PulsePattern2_RET       ret

'---------------- Uninitialized data ----------------

'32-bit constants
'cog instructions can not handle literals beyond (511), 2^9-1, as
'their internal source and destination fields are only 9-bits in width.

cog_true      long      TRUE        '$FFFFFFFF  (-1)
cog_false     long      FALSE       '$00000000  (0)

TestMask      long      |<TestPin   

'temp general
t1                      res     1
t2                      res     1
t3                      res     1

'temp address
a1                      res     1
a2                      res     1
'temp data
d1                      res     1
d2                      res     1
'temp counter
c1                      res     1
c2                      res     1

'temp for self-modifying code
xx                      res     1

'pin masks
OutMask                 res     1
Out_LeadEdge0           res     1
Out_LeadEdge1           res     1
TrigMask                res     1
AltMask                 res     1
GateMask                res     1
TrigGateMask            res     1

'other variables
SyncPoint               res     1
Initialized             res     1
PulseSetup              res     1
To_end_of_period        res     1
AltFlag                 res     1

'******* ORDER OF FOLLOWING VARIABLES IS CRITICAL *******

'Arbitrary convention:  all cog variables to be copied from hub-to-cog 
'                       will be prefixed with an underscore character. 

'Cog copy of Hub _index                      
a_index                 res     1

'Parameter list, ParaCnt == 7
a_mode                  res     1   'parameter #1
a_delay                 res     1   '#2
a_width                 res     1   '#3
a_period                res     1   '#4
a_number                res     1   '#5
a_sync                  res     1   '#6
a_altpin                res     1   '#7
a_gatepin               res     1   '#8

'Read-only list (only accessible via Read_Parameter() method)
'WARNING: critical that this variable immediately follow last parameter in cog memory
PulseCounter            res     1   '#ParaCnt+1  'read-only



                        FIT    'generate a compile-time error if assem won't fit in cog